﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;

namespace Delay
{
    /// <summary>
    /// Class that encodes a sequence of pixels to a sequence of bytes representing a PNG file (uncompressed).
    /// </summary>
    /// <remarks>
    /// Reference: http://www.w3.org/TR/PNG/.
    /// </remarks>
    static class PngEncoder
    {
        /// <summary>
        /// Maximum length of an uncompressed block.
        /// </summary>
        private const int MaximumUncompressedBlockLength = 65535;

        /// <summary>
        /// Encodes the specified pixels to a sequence of bytes representing a PNG file.
        /// </summary>
        /// <param name="width">Width of the image.</param>
        /// <param name="height">Height of the image.</param>
        /// <param name="pixels">Pixels of the image in ARGB format.</param>
        /// <returns>Sequence of bytes representing a PNG file.</returns>
        public static IEnumerable<byte> Encode(int width, int height, IEnumerable<int> pixels)
        {
            if (null == pixels)
            {
                throw new ArgumentNullException("pixels");
            }
            if (width <= 0)
            {
                throw new ArgumentException("Width must be positive.", "width");
            }
            if (height <= 0)
            {
                throw new ArgumentException("Height must be positive.", "height");
            }

            return Concat(
                new byte[] { 137, 80, 78, 71, 13, 10, 26, 10 }, // PNG signature
                GetIhdrChunk(width, height),                    // IHDR
                GetIdatChunk(width, height, pixels),            // IDAT
                GetIendChunk());                                // IEND
        }

        /// <summary>
        /// Gets a sequence of bytes representing the IHDR chunk.
        /// </summary>
        /// <param name="width">Width of the image.</param>
        /// <param name="height">Height of the image.</param>
        /// <returns>Sequence of bytes representing the IHDR chunk.</returns>
        private static IEnumerable<byte> GetIhdrChunk(int width, int height)
        {
            var ihdrData = Concat(
                BitConverter.GetBytes(width).BigEndian(),
                BitConverter.GetBytes(height).BigEndian(),
                new byte[]
                {
                    8, // 8-bit color depth
                    6, // Truecolor with alpha (RGBA)
                    0, // Deflate compression
                    0, // Adaptive (no) filtering
                    0, // No interlace
                });

            return CreateChunk(
                new byte[] { 73, 72, 68, 82 }, // IHDR
                ihdrData,
                4 + 4 + 5); // Width + Height + Header
        }

        /// <summary>
        /// Gets a sequence of bytes representing the IDAT chunk.
        /// </summary>
        /// <param name="width">Width of the image.</param>
        /// <param name="height">Height of the image.</param>
        /// <param name="pixels">Pixels of the image in ARGB format.</param>
        /// <returns>Sequence of bytes representing the IDAT chunk.</returns>
        private static IEnumerable<byte> GetIdatChunk(int width, int height, IEnumerable<int> pixels)
        {
            using (var uncompressedBytes = new Adler32WrappedEnumerable(GetFilteredScanlines(width, height, pixels)))
            {
                var uncompressedByteCount = (width * height * 4) + height; // 4 bytes per pixel, one filter byte per line

                var idatData = Concat(
                    new byte[] { 0x78, 0x01 }, // deflate compression, 32k window size, no preset dictionary, minimum (no) compression
                    GetUncompressedBlocks(uncompressedBytes, uncompressedByteCount),
                    uncompressedBytes.Hash);   // Adler-32
                var idatDataCount =
                    2 +                     // Header
                    uncompressedByteCount + // Uncompressed byte count
                    4 +                     // Adler32 hash
                    (5 * ((uncompressedByteCount + (MaximumUncompressedBlockLength - 1)) / MaximumUncompressedBlockLength)); // 5 bytes of compressed block header for every 65535 bytes

                return CreateChunk(
                    new byte[] { 73, 68, 65, 84 }, // IDAT
                    idatData,
                    idatDataCount);
            }
        }

        /// <summary>
        /// Gets a sequence of bytes representing the IEND chunk.
        /// </summary>
        /// <returns>Sequence of bytes representing the IEND chunk.</returns>
        private static IEnumerable<byte> GetIendChunk()
        {
            return CreateChunk(
                new byte[] { 73, 69, 78, 68 }, // IEND
                Enumerable.Empty<byte>(),
                0); // No content
        }

        /// <summary>
        /// Gets the filtered scanlines for the specified image pixels.
        /// </summary>
        /// <param name="width">Width of the image.</param>
        /// <param name="height">Height of the image.</param>
        /// <param name="pixels">Pixels of the image in ARGB format.</param>
        /// <returns>Sequence of bytes representing the filtered scanlines.</returns>
        private static IEnumerable<byte> GetFilteredScanlines(int width, int height, IEnumerable<int> pixels)
        {
            var x = 0;
            var y = 0;
            foreach (var p in pixels)
            {
                if (0 == x)
                {
                    // New scanline; output filter byte
                    yield return 0; // Filter type 0 (direct mapping)
                }
                yield return (byte)(p >> 16); // R
                yield return (byte)(p >> 8);  // G
                yield return (byte)(p >> 0);  // B
                yield return (byte)(p >> 24); // A
                x++;
                if (width == x)
                {
                    // Scanline complete; next row
                    x = 0;
                    y++;
                    if (height == y)
                    {
                        // All scanlines complete; done
                        break;
                    }
                }
            }
            if (height != y)
            {
                throw new InvalidOperationException("Not enough pixel data for specified width and height.");
            }
        }

        /// <summary>
        /// Gets uncompressed blocks (per the deflate algorithm) for the specified bytes.
        /// </summary>
        /// <param name="bytes">Bytes to compress.</param>
        /// <param name="count">Count of bytes.</param>
        /// <returns>Sequence of bytes representing the uncompressed blocks.</returns>
        private static IEnumerable<byte> GetUncompressedBlocks(IEnumerable<byte> bytes, int count)
        {
            var bytesInBlock = 0;
            foreach (var b in bytes)
            {
                if (0 == bytesInBlock)
                {
                    // New compressed block; output header
                    var blockLength = Math.Min(count, MaximumUncompressedBlockLength);
                    var bfinal = new byte[] { (blockLength < count) ? (byte)0x00 : (byte)0x01 };
                    var len = BitConverter.GetBytes((ushort)blockLength).LittleEndian();
                    var nlen = BitConverter.GetBytes((ushort)~blockLength).LittleEndian();
                    var header = Concat(
                        bfinal, // BFINAL bit
                        len,    // Length
                        nlen);  // 1's complement of length

                    // Output header bytes
                    foreach (var h in header)
                    {
                        yield return h;
                    }
                }
                // Output byte
                yield return b;
                bytesInBlock++;
                if (MaximumUncompressedBlockLength == bytesInBlock)
                {
                    // Block complete; prepare for next block
                    count -= bytesInBlock;
                    bytesInBlock = 0;
                }
            }
        }

        /// <summary>
        /// Creates a PNG chunk of the specified type, data, and length.
        /// </summary>
        /// <param name="chunkType">Chunk type.</param>
        /// <param name="chunkData">Chunk data.</param>
        /// <param name="count">Count of data bytes.</param>
        /// <returns>Sequence of bytes representing the PNG chunk.</returns>
        private static IEnumerable<byte> CreateChunk(byte[] chunkType, IEnumerable<byte> chunkData, int count)
        {
            using (var bytes = new Crc32WrappedEnumerable(Concat(chunkType, chunkData)))
            {
                return Concat(
                    BitConverter.GetBytes(count).BigEndian(), // Type
                    bytes,                                    // Data
                    bytes.Hash);                              // CRC-32
            }
        }

        /// <summary>
        /// Concatenates an arbitrary number of enumerations into a single one.
        /// </summary>
        /// <typeparam name="T">Type of the enumerations.</typeparam>
        /// <param name="args">Array of enumerations to concatenate.</param>
        /// <returns>Sequence representing the concatenated enumerations.</returns>
        private static IEnumerable<T> Concat<T>(params IEnumerable<T>[] args)
        {
            var concat = Enumerable.Empty<T>();
            foreach (var arg in args)
            {
                concat = Enumerable.Concat(concat, arg);
            }
            return concat;
        }

        /// <summary>
        /// Returns the little-endian (LSB first) representation of the specified enumeration.
        /// </summary>
        /// <typeparam name="T">Type of the enumeration.</typeparam>
        /// <param name="enumeration">Enumeration.</param>
        /// <returns>Little-endian representation of the enumeration.</returns>
        internal static IEnumerable<T> LittleEndian<T>(this IEnumerable<T> enumeration)
        {
            return BitConverter.IsLittleEndian ? enumeration : enumeration.Reverse();
        }

        /// <summary>
        /// Returns the big-endian (MSB first) representation of the specified enumeration.
        /// </summary>
        /// <typeparam name="T">Type of the enumeration.</typeparam>
        /// <param name="enumeration">Enumeration.</param>
        /// <returns>Big-endian representation of the enumeration.</returns>
        internal static IEnumerable<T> BigEndian<T>(this IEnumerable<T> enumeration)
        {
            return BitConverter.IsLittleEndian ? enumeration.Reverse() : enumeration;
        }
    }

    /// <summary>
    /// Class that wraps an IEnumerable(T) or IEnumerator(T) in order to do something with each byte as it is enumerated.
    /// </summary>
    /// <typeparam name="T">Type of element being enumerated.</typeparam>
    abstract class WrappedEnumerable<T> : IEnumerable<T>, IEnumerator<T>
    {
        /// <summary>
        /// Wrapped enumerator.
        /// </summary>
        private readonly IEnumerator<T> _enumerator;

        /// <summary>
        /// True iff initialized.
        /// </summary>
        private bool _initialized;

        /// <summary>
        /// Method called to initialize for a new (or reset) enumeration.
        /// </summary>
        protected virtual void Initialize() { }

        /// <summary>
        /// Method called for each byte output by the underlying enumerator.
        /// </summary>
        /// <param name="value">Next value.</param>
        protected abstract void OnValueEnumerated(T value);

        /// <summary>
        /// Initializes a new instance of the WrappedEnumerable class.
        /// </summary>
        /// <param name="enumerable">Underlying enumeration that is being wrapped.</param>
        public WrappedEnumerable(IEnumerable<T> enumerable)
        {
            _enumerator = enumerable.GetEnumerator();
        }

        /// <summary>
        /// Initializes a new instance of the WrappedEnumerable class.
        /// </summary>
        /// <param name="enumerable">Underlying enumerator that is being wrapped.</param>
        public WrappedEnumerable(IEnumerator<T> enumerator)
        {
            _enumerator = enumerator;
        }

        /// <summary>
        /// Gets an enumerator.
        /// </summary>
        /// <returns>Enumerator instance.</returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this;
        }

        /// <summary>
        /// Gets an enumerator.
        /// </summary>
        /// <returns>Enumerator instance.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        /// <summary>
        /// Gets the current element of the enumeration.
        /// </summary>
        public T Current
        {
            get
            {
                if (!_initialized)
                {
                    Initialize();
                    _initialized = true;
                }
                T value = _enumerator.Current;
                OnValueEnumerated(value);
                return value;
            }
        }

        /// <summary>
        /// Gets the current element of the enumeration.
        /// </summary>
        object IEnumerator.Current
        {
            get { return Current; }
        }

        /// <summary>
        /// Moves to the next element of the enumeration.
        /// </summary>
        /// <returns>True iff another element is available.</returns>
        public bool MoveNext()
        {
            return _enumerator.MoveNext();
        }

        /// <summary>
        /// Resets the enumeration.
        /// </summary>
        public void Reset()
        {
            _enumerator.Reset();
            _initialized = false;
        }

        /// <summary>
        /// Disposes of any resources.
        /// </summary>
        public void Dispose()
        {
            _enumerator.Dispose();
        }
    }

    /// <summary>
    /// Class that implements the Adler-32 algorithm by wrapping a sequence of bytes and hashing them on the fly.
    /// </summary>
    /// <remarks>
    /// Reference: http://www.ietf.org/rfc/rfc1950.txt.
    /// </remarks>
    class Adler32WrappedEnumerable : WrappedEnumerable<byte>
    {
        /// <summary>
        /// Modulo constant.
        /// </summary>
        private const int AdlerModulo = 65521;

        /// <summary>
        /// State variable 1.
        /// </summary>
        private int _s1;

        /// <summary>
        /// State variable 2.
        /// </summary>
        private int _s2;

        /// <summary>
        /// Initializes a new instance of the Crc32WrappedEnumerable class.
        /// </summary>
        /// <param name="enumerable">Sequence to compute the hash for.</param>
        public Adler32WrappedEnumerable(IEnumerable<byte> enumerable)
            : base(enumerable) { }

        /// <summary>
        /// Initializes the hash algorithm.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
            _s1 = 1;
            _s2 = 0;
        }

        /// <summary>
        /// Processes the next byte and updates the hash.
        /// </summary>
        /// <param name="value">Next byte.</param>
        protected override void OnValueEnumerated(byte value)
        {
            _s1 = (_s1 + value) % AdlerModulo;
            _s2 = (_s2 + _s1) % AdlerModulo;
        }

        /// <summary>
        /// Returns the hash in big-endian byte order.
        /// </summary>
        public IEnumerable<byte> Hash
        {
            get
            {
                foreach (var b in BitConverter.GetBytes((_s2 << 16) | _s1).BigEndian())
                {
                    yield return b;
                }
            }
        }
    }

    /// <summary>
    /// Class that implements the CRC-32 algorithm by wrapping a sequence of bytes and hashing them on the fly.
    /// </summary>
    /// <remarks>
    /// Reference: http://www.w3.org/TR/PNG/.
    /// </remarks>
    class Crc32WrappedEnumerable : WrappedEnumerable<byte>
    {
        /// <summary>
        /// Shared, pre-computed lookup table for efficiency.
        /// </summary>
        private static readonly uint[] _crc32Table;

        /// <summary>
        /// Class constructor initializes the shared lookup table.
        /// </summary>
        [SuppressMessage("Microsoft.Performance", "CA1810:InitializeReferenceTypeStaticFieldsInline", Justification = "Table values must be computed; not possible to remove the static constructor.")]
        static Crc32WrappedEnumerable()
        {
            // Allocate table
            _crc32Table = new uint[256];

            // For each byte
            for (uint n = 0; n < 256; n++)
            {
                // For each bit
                uint c = n;
                for (int k = 0; k < 8; k++)
                {
                    // Compute value
                    if (0 != (c & 1))
                    {
                        c = 0xedb88320 ^ (c >> 1);
                    }
                    else
                    {
                        c = c >> 1;
                    }
                }

                // Store result in table
                _crc32Table[n] = c;
            }
        }

        /// <summary>
        /// Current hash value.
        /// </summary>
        private uint _crc32Value;

        /// <summary>
        /// Initializes a new instance of the Crc32WrappedEnumerable class.
        /// </summary>
        /// <param name="enumerable">Sequence to compute the hash for.</param>
        public Crc32WrappedEnumerable(IEnumerable<byte> enumerable)
            : base(enumerable) { }

        /// <summary>
        /// Initializes the hash algorithm.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
            _crc32Value = uint.MaxValue;
        }

        /// <summary>
        /// Processes the next byte and updates the hash.
        /// </summary>
        /// <param name="value">Next byte.</param>
        protected override void OnValueEnumerated(byte value)
        {
            byte index = (byte)(_crc32Value ^ value);
            _crc32Value = _crc32Table[index] ^ ((_crc32Value >> 8) & 0xffffff);
        }

        /// <summary>
        /// Returns the hash in big-endian byte order.
        /// </summary>
        public IEnumerable<byte> Hash
        {
            get
            {
                foreach (var b in BitConverter.GetBytes(~_crc32Value).BigEndian())
                {
                    yield return b;
                }
            }
        }
    }
}
